/*  
**  File: pset.h
**  Purpose: header file for the PSET package
**  Author: Song Mao (maosong@cfar.umd.edu)
**  Date:   16, September 2000
**  Organization: Language and Media Processing Lab
**                 Center for Automation Research
**                   University of Maryland
**                   College Park, MD 20742
**  References:
**  $Id: pset.h,v 1.1 2000/09/19 22:20:12 maosong Exp maosong $
**
*/

#include <stdio.h>
#include "rafport.h"
#include "dafstype.h"
#include <sys/types.h>
#include <sys/uio.h>
#include "rlimage.h"
#include "dafs.h"

#define	YES	1
#define NO	0
#define NO_TYPE  -1
#define ENCLOSE   1
#define SPLIT     2
#define APART     3
#define MERGE     4
#define MAX_SG  100


typedef struct dataset {        /* dataset data structure */
        int nImg;               /* total number of images in the dataset */
        char **ImgName;         /* image names in the dataset */
} DATASET;

typedef struct train_protocol {
        DATASET         *dataset;            /* training dataset */
        char            *gt_dir;             /* ground truth file directory */        
        char            *img_dir;               /*  image file directory */
        char            *gt_suffix;             /* ground truth file suffix */
        char            *sg_suffix;             /* segmentation result file suffix */
        char            *img_suffix;            /* image file suffix */
        char            *train_result_dir;      /* training result file directory */
        char            opt_alg;                /* optimization algorithm */
        char            ben_alg;                /* benchmarking algorithm */
        char            seg_alg;                /* segmentation algorithm */
} TRAIN_PROTOCOL;

typedef struct test_protocol {
        DATASET         *dataset;            /* testing dataset */
        char            *gt_dir;                /* ground truth file directory */        
        char            *img_dir;               /*  image file directory */
        char            *gt_suffix;             /* ground truth file suffix */
        char            *sg_suffix;             /* segmentation result file suffix */
        char            *img_suffix;            /* image file suffix */
        char            *test_result_dir;      /* testing result file directory */
        char            ben_alg;                /* benchmarking algorithm */
        char            seg_alg;                /* segmentation algorithm */
} TEST_PROTOCOL;

typedef struct bench_param {    /* benchmark algorithm parameter data structure */
        int     htol;           /* horizontal overlap tolerance in percentage */
        int     vtol;           /* vertical overlap tolerance in percentage */
        int     hpix;           /* horizontal overlap tolerance in number of pixels */
        int     vpix;           /* vertical overlap tolerance in number of pixels */
} BENCH_PARAM;

typedef struct weight {		/* weight data structure */
        int     wSpl;           /* weight for number of splits */
        int     wMrg;           /* weight for number of horizontal merges */
        int     wMis;           /* weight for number of mis-detections */
        int     wFA;            /* weight for number of false alarms */
        int     wSplLine;       /* weight for number of split textlines */
        int     wMrgLine;       /* weight for number of horizontally merged textlines */
        int     wMisLine;       /* weight for number of mis-detected textlines */
        int     wFAZone;        /* weight for number of false alarm regions */
} WEIGHT;

typedef struct textline{
  	int SG_ID[MAX_SG];
  	int SG_TYPE[MAX_SG];
  	int sg_index;
  	iBox box;
  	int miss_det;
  	int h_split;
  	int h_merge;
  	int v_split;
  	int v_merge;
  	int textline_flg;
} TEXTLINE;


typedef struct zone{
  	iBox box;
  	int false_alarm;
  	int line_counter;
} ZONE;

typedef struct err_measure {
        int     nSpl;
        int     nMrg;
        int     nMis;
        int     nFA;
        int     nSplLine;
        int     nMrgLine;
        int     nMisLine;
        int     nFAZone;
        int     nErrLine;
        int     nGtLine;
} ERR_MEASURE;

typedef struct score {          /* score data structure */
        float err_rate;         /* segmentation algorithm error rate */
        float timing;           /* segmentation algorithm timing */
} SCORE;

/**---------------------function bodys---------------------------**/
void ReadInput(char *gtdafs, char *sgdafs, iEntityPtr *gt_doc_p,
	iEntityPtr *sg_doc_p, iDAFSError *dafs_error_p);
void ReadZoneInfo(iEntityPtr gt_doc, iEntityPtr sg_doc, int *gt_zone_count_p,
	  	int *sg_zone_count_p, int **textline_count_p,
		TEXTLINE  ***textline_p, iBox **gt_zone_p, ZONE **sg_zone_p);
void GetZoneNo(iEntityPtr root, int *gt_zone_count_p);
int GetTextlineInfo(iEntityPtr root, int *gt_zone_count_p, iBox **gt_zone_p,
	int **textline_count_p, TEXTLINE ***textline_p);
void GetSGZoneNo(iEntityPtr root, int *sg_zone_count_p);
int GetSGZoneInfo(iEntityPtr root, int *sg_zone_count_p, ZONE **sg_zone_p);
void CmpGt_Sg(BENCH_PARAM *para, char *gtdafs, TEXTLINE **textline,
		iBox *gt_zone, ZONE *sg_zone, int gt_zone_count,
		int sg_zone_count, int *textline_count, ERR_MEASURE **err_measure_p);
int DetectSplit(BENCH_PARAM *para, TEXTLINE ***textline_p,
		ZONE **sg_zone_p, int gt_zone_count,
		int sg_zone_count, int *textline_count);
int DetectMerge(BENCH_PARAM *para, TEXTLINE ***textline_p,
		iBox *gt_zone, ZONE **sg_zone_p, int gt_zone_count,
		int sg_zone_count, int *textline_count);
int DetectMissDet(BENCH_PARAM *para, TEXTLINE ***textline_p,
		int *textline_count, int gt_zone_count);
int DetectFalseAlarm(BENCH_PARAM *para, ZONE *sg_zone, int sg_zone_count);
double CmpScore();
int UnionOf(BENCH_PARAM *para, ZONE a, TEXTLINE b);
void GetError(char *img, TEXTLINE **textline, iBox *gt_zone, ZONE *sg_zone,
		int gt_zone_count, int sg_zone_count, int *textline_count);

void ReadWeight(char *weight_file, WEIGHT **weight_p);

/*
**  Docstrum Segmentation Algorithm replated header files
**
*/

#define MAX_RECTS 20000
#define MAX_COMPS 20000
#define FUNC_CALL       2       /* Call segmentation algorithm by its function */
#define SHELL_CALL      3       /* Call segmentation algorithm by its executable */

typedef struct docstrum_param{	/* Docstrum algorithm parameter data structure */
        char    alg_mode;       /* algorithm call mode */
	char    *concom_dir;	/* connected component file directory */
	char	*concom_suffix; /* connected component file suffix */
        int     knn;        	/* number of nearest connected components to be clustered */
        int     lsize;      	/* low size threshold for connected component */
        int     hsize;      	/* high size threshold for connected component */
        int     hatol;      	/* horizontal angle tolerance */
        int     vatol;      	/* vertical angle tolerance */
        float   tcc;        /* threshold for clustering connected component into lines */
        float   par;        /* parallel threshold for clustering lines into zones */
        float   per;        /* perpendicular threshold for clustering lines into zones */
        float   scr;        /* super/sub script distance threshold */
        float   dis;        /* connected component size grouping threshold */
} DOCSTRUM_PARAM;

typedef struct est_para {
  	double or;
  	int cs;
  	int ts;
} EST_PARA;

typedef struct dis {
  	double d;
  	double a;
  	int    i;
} DIS;

 
typedef struct node_docstrum
{
	int v;
  	struct node_docstrum *next;
} NODE_DOCSTRUM;

typedef struct rect{ /* text-line, zone or document look-up table */
  	int x1;
  	int y1;
  	int x2;
  	int y2;
  	int flg;
  	int *child;
} RECT_DOCSTRUM;




/*-----------functions definations------------*/
float dis(RECT_DOCSTRUM a, RECT_DOCSTRUM b);
float ang(RECT_DOCSTRUM a, RECT_DOCSTRUM b);
int par_dist(RECT_DOCSTRUM a, RECT_DOCSTRUM b);
int per_dist(RECT_DOCSTRUM a, RECT_DOCSTRUM b);

int ReadTIFFImageDocstrum(char *tif_filename, RECT_DOCSTRUM **document_p,
	RECT_DOCSTRUM **line_p, RECT_DOCSTRUM **zone_p, DOCSTRUM_PARAM **para_p);
void ReadCC(char *cc_filename, DOCSTRUM_PARAM *para, int *cc_index_p,
	int *cc_index1_p, int *cc_height_p, int *cc_height1_p,
	RECT_DOCSTRUM **comp_p, RECT_DOCSTRUM **comp1_p, DIS ***d_p,
	int *rerun_p, float *height_ratio_p);
void EstPara(int runs, RECT_DOCSTRUM *document, RECT_DOCSTRUM *comp,
	DIS ***d_p, int cc_index, int cc_height, float height_ratio,
	DOCSTRUM_PARAM *para, EST_PARA *est_para_p, int *line_t_p,
	int *par_t_p, int *per_t_p, int *par_t_h_p, int *per_t_h_p);
void MakeTextline(EST_PARA est_para, DOCSTRUM_PARAM *para,
	RECT_DOCSTRUM *comp, DIS **d, int cc_index, int line_t,
	int *line_index_p, RECT_DOCSTRUM **line_p);
void MakeZone(EST_PARA est_para, DOCSTRUM_PARAM *para,
	RECT_DOCSTRUM *line, int line_index, int par_t, int per_t,
	int par_t_h, int per_t_h, int *zone_index_p, RECT_DOCSTRUM **zone_p);
DIS **dis_matrix(int nrl, int nrh, int ncl, int nch);
void free_dis_matrix(DIS **m, int nrl, int nrh, int ncl, int nch);
void TransitiveClosure(int E, int *x, int *y, RECT_DOCSTRUM *rect1,
	int rect1_index_s, int rect1_index_e, int *rect2_index_p,
	RECT_DOCSTRUM **rect2_p);
void visit(int k, RECT_DOCSTRUM *rect1, NODE_DOCSTRUM *z, NODE_DOCSTRUM ***adj_p,
	int *id_p, int **val_p, int *index_p, RECT_DOCSTRUM **rect2_p, int *ch_index_p);
void FreeJobDocstrum(RECT_DOCSTRUM *zone, RECT_DOCSTRUM *document);
void FreeJob1(DIS **d, int cc_index);
void GetDAFS(char *tif_filename, char *dafs_filename, int line_index,
	int zone_index, RECT_DOCSTRUM *document, RECT_DOCSTRUM *line,
	RECT_DOCSTRUM *zone);
void sort2(int n, float *comp_x, int *comp_i);
void selectK(float a[], int b[], int n, int k);

/*
**  Voronoi-based segmentation algorithm related header files
**
*/

#define FUNC_CALL       2     /* Call segmentation algorithm by its function */
#define SHELL_CALL      3     /* Call segmentation algorithm by its executable*/

typedef struct voronoi_param{
        char    alg_mode;       /* algorithm call mode */
  	int 	sr;
  	int 	nm;
  	float 	fr;
  	int 	ta;
  	int 	sw;
  	int 	Cw;
  	int 	Ch;
  	int 	Cr;
  	int 	Az;
  	int 	Al;
  	int 	Br;
} VORONOI_PARAM;

void VoronoiFunction();

/*
**  The X-Y cut segmentation algorithm related header files.
**
*/

#define IN_INCH 1
#define MAX 100000
#define STOP_HERE 0
#define GO_ON     1
#define HORIZONTAL_CUT 0
#define VERTICAL_CUT   1

#define FUNC_CALL       2       /* Call segmentation algorithm by its function */
#define SHELL_CALL      3       /* Call segmentation algorithm by its executable */

typedef struct xycut_param { /* X-Y cut segmentation algorithm parameter data structure */
	char    alg_mode;	/* algorithm call mode */
        int     Tnx;            /* X noise removal threshold */
        int     Tny;            /* Y noise removal threshold */
        int     Tcx;            /* X cut threshold */
        int     Tcy;            /* Y cut threshold */
} XYCUT_PARAM;

typedef struct node_xycut{
  	struct node_xycut *l, *r;
  	char *info;
  	int ul_x;
  	int ul_y;
  	int width;
  	int height;
  	int *x_histogram;
  	int *y_histogram;
  	int local_Tnx;
  	int local_Tny;
}NODE_XYCUT; 

typedef struct {
  	int width;
  	int height;
  	int depth;
  	int xres;
  	int yres;
  	char *image;
  	int rowsize;
} DOCUMENT_IMAGE;

/****function declaration*****/
int ReadTIFFImageXycut( char *infile, XYCUT_PARAM **para_p,
	DOCUMENT_IMAGE **document, NODE_XYCUT **root, char ***buf);
void Create_Histogram_Table(char **buf, DOCUMENT_IMAGE *document,
	int ***x_hist_table_p, int ***y_hist_table_p);
void ZoneAdjust(char **buf, NODE_XYCUT *root, DOCUMENT_IMAGE *document, 
		 int **x_hist_table, int **y_hist_table,
                 NODE_XYCUT **zone_p, int ***free_hist_p, int *h_p);
int GetRdiff(NODE_XYCUT *root, char *intif, char *rdifffile, DOCUMENT_IMAGE *document);
int WriteTIFFImage(DOCUMENT_IMAGE *document,int	resunit, float xresolution,
		   float yresolution, char *description, char *filename,
		   int  *ierr);
void GetHistogram(NODE_XYCUT **t, int **x_hist_table, int **y_hist_table,
		DOCUMENT_IMAGE *document);
int SplitZone(char **buf, NODE_XYCUT *root, NODE_XYCUT **zone_p, 
	DOCUMENT_IMAGE *document, int **x_hist_table, int **y_hist_table,
	XYCUT_PARAM *para, int *cut_direction_p, int *updated_leftwidth_p,
	int *updated_upperheight_p, int ***free_hist_p, int *h_p);
int MakeTree(char *tif_filename, char **buf, NODE_XYCUT *root,
	DOCUMENT_IMAGE *document, int **x_hist_table, int **y_hist_table,
	XYCUT_PARAM *para, char *dafs_filename); 
int push(NODE_XYCUT *p, int *sp_p, NODE_XYCUT ***stack_p);
NODE_XYCUT * pop(int *sp_p, NODE_XYCUT ***stack_p, NODE_XYCUT **end_node);
int PushLeaf(NODE_XYCUT *p, int *leaf_sp_p, NODE_XYCUT ***leaf_stack_p);
NODE_XYCUT * PopLeaf(int *leaf_sp_p, NODE_XYCUT ***leaf_stack_p, NODE_XYCUT **end_node_p);
void FreeJobXycut(DOCUMENT_IMAGE *document, XYCUT_PARAM *para, char **buf,
	int **x_hist_table, int **y_hist_table);

/*
**  GenscoreDocstrum related header files
**
*/

int DocstrumFunction(char *tif_file, char *cc_file,
	DOCSTRUM_PARAM *para, char *dafs_file);
void Bench(char *img, BENCH_PARAM *para, char *gtdafs, char *sgdafs,
           ERR_MEASURE **err_measure_p);
float BenchScoring(BENCH_PARAM *para, ERR_MEASURE *err, WEIGHT *weight);

/* for sh_c */
void Usage(void);
/* extern char* progname;  */

/*
**  GenscoreVoronoi related header files.
**
*/

void VoronoiFunction(char *tif_filename, VORONOI_PARAM *para, char *dafs_filename);

/*
**  GenscoreXycut related header files
*/

int XycutFunction(char *tif_filename, XYCUT_PARAM *para, char *dafs_filename);

/*
**  Simplex optimization algorithm related header file
*/

#define NM              0       /* Nelder-Mead stop criterion */
#define NR              1       /* Numerical Recipe stop criterion */
#define FUNC_CALL       2       /* Call segmentation algorithm by its function */
#define SHELL_CALL      3       /* Call segmentation algorithm by its executable */
#define ON      	1       
#define OFF      	0      

typedef struct simplex_param {   /* simplex algorithm parameter data structure */
	int     ndim;           /* variable dimension of the objective function */
        int     criflg;         /* stop criterion flag */
        int     nmax;           /* maximum number of function evaluations allowed */
        float   ftol;           /* function difference tolerance */
        float   alpha;          /* simplex reflection coefficient */
        float   beta;           /* simplex contraction coefficient */
        float   gamma;          /* simplex expansion coefficient */
        float   sigma;          /* simplex shrinkage coefficient */
        float   **p;            /* objective function variable vector */
        float   *y;             /* objective function value vector */
        float   *t;             /* timing information vector */
        float   *scale;         /* initial simplex scale vector */
} SIMPLEX_PARAM;

#define GET_PSUM for(j = 1; j <= opt_param->ndim; j++) { \
                   for(i = 1, sum = 0.0; i <= mpts; i++) \
                     sum += opt_param->p[i][j]; psum[j] = sum;}

/*
**  Train module related header file
*/

void GenscoreXycut(TRAIN_PROTOCOL *train_protocol, void *xycut_param,
                     void *bench_param, char *weight_file, float *val, SCORE **score_p);
void GenscoreDocstrum(TRAIN_PROTOCOL *train_protocol, void *docstrum_param,
                       void *bench_param, char *weight_file, float *val, SCORE **score_p);
void GenscoreVoronoi(TRAIN_PROTOCOL *train_protocol, void *voronoi_param,
                     void *bench_param, char *weight_file, float *val, SCORE **score_p);
void GenscoreTestDriver(TRAIN_PROTOCOL *train_protocol, void *seg_param,
                        void *bench_param, char *weight_file, float *val, SCORE **score_p);

/*
**  Test module related header file
*/


void EvalXycut();
void EvalDocstrum();
void EvalVoronoi();

void ReadFileNames(char *datafile, DATASET **dataset_p);

/*
**  TrainSeg module related header file
*/

/* Segmentation Algorithms */
#define XYCUT           6       /* X-Y cut segmentation algorithm by Nagy */
#define DOCSTRUM        7       /* Docstrum segmentation algorithm by O'Gorman */
#define VORONOI         8       /* Voronoi-based segmentation algorithm by Kise */
#define TEST_DRIVER     11      /* Test driver for optimization verification */

/* Optimization Algorithms */
#define SIMPLEX         9       /* simplex optimization algorithm by Nelder-Mead */

/* Benchmarking Algorithms */
#define TEXTLINE_BASED  10


char * CalenderTime();
/* function declarations */
void ParseArgTrain(int argc, char **argv,
        char **train_report_p, char **opt_seg_param_p,
	char **train_protocol_file_p, char **opt_file_p,
	char **seg_file_p, char **ben_file_p, char **weight_file_p);
void ReadTrainProtocol(char *train_protocol_file, TRAIN_PROTOCOL **train_protocol_p);
void ReadSimplexParam(char *simplex_file, void **simplex_param_p);
void ReadBenchParam(char *bench_file, void **ben_param_p);
void ReadXycutParam(char *xycut_file, void **xycut_param_p);
void ReadDocstrumParam(char *docstrum_file, void **docstrum_param_p);
void ReadVoronoiParam(char *voronoi_file, void **voronoi_param_p);
void WriteTrainReportHeader(int argc, char **argv, TRAIN_PROTOCOL *train_protocol, char **file_p);
void Train(TRAIN_PROTOCOL *train_protocol, void *opt_param,
           void *ben_param, void *seg_param, char *weight_file,
	   char **train_report_p, char **opt_seg_param_p);
void FreeJobTrain(TRAIN_PROTOCOL *train_protocol, char *train_report,
		  char *opt_seg_param, char *train_protocol_file,
		  void *opt_param, void *ben_param,
                  void *seg_param, char *opt_file, char *seg_file,
                  char *ben_file, char *weight_file);

/*
**  TestSeg module related header file
**  Purpose: header file for the main function
**  Author: Song Mao (maosong@cfar.umd.edu)
**  Date:   29, March 1999
**  Organization: Language and Media Processing Lab
**                 Center for Automation Research
**                   University of Maryland
**                   College Park, MD 20742
**  $Id: pset.h,v 1.1 2000/09/19 22:20:12 maosong Exp maosong $
**
*/

/* function declarations */
void ParseArgTest();
void ReadTestProtocol();
void WriteTestReportHeader();
void Test();
void FreeJobTest();


float Amotry(TRAIN_PROTOCOL *train_protocol, void *seg_param, void *ben_param,
             char *weight_file, SIMPLEX_PARAM *simplex_param, float *psum,
             void (*genscore)(TRAIN_PROTOCOL *, void *, void *, char *,
                float *, SCORE **),
             int ihi, int *nFuncEval_p, float fac, FILE *fp);

int Simplex(TRAIN_PROTOCOL *train_protocol, SIMPLEX_PARAM *opt_param,
            void *ben_param, void *seg_param, char *weight_file,
            void (*genscore)(TRAIN_PROTOCOL *, void *, void *, char *,
		float *, SCORE **),
            char **train_report_p, char **opt_seg_param_p);

/* Voronoi-based algorithm related header files */

/*
   $Id: pset.h,v 1.1 2000/09/19 22:20:12 maosong Exp maosong $
   $Date: 2000/09/19 22:20:12 $
   $Revision: 1.1 $
   $Author: maosong $
*/

#ifndef DEFS_H_INCLUDED_
#define DEFS_H_INCLUDED_

/*
   $Id: pset.h,v 1.1 2000/09/19 22:20:12 maosong Exp maosong $
   $Date: 2000/09/19 22:20:12 $
   $Revision: 1.1 $
   $Author: maosong $
*/
#ifndef NULL
#define NULL	0
#endif
#define DELETED		-2

#define	BYTE		8       /* [bit] */
#define	BLACK		1
#define WHITE		0
#define	NAMELEN		500

/* initial size and increments */
#define INITPIXEL	5000000  /* initial size of black pixels */
#define INCPIXEL        100000   /* increments of black pixels */
#define INITNEIGHBOR	50000    /* initial size of the array "neighbor" */
#define INCNEIGHBOR     10000    /* its increments */

#define INITLINE        50000    /* initial size of the array "lineseg" */
#define INCLINE         10000    /* its increments */
#define SITE_BOX        10000
#ifdef RECT
#define INITCRECT       50000
#define INCCRECT        10000
#define INITBRECT       500
#define INCBRECT        100
/*
#define INITLABELLIST   500
#define INCLABELLIST    100
*/
#endif /* RECT*/

/* MAX valures */
#define UNSIGNED_MAX    0xFFFFFFFF /* max. of unsigned int */
#define NOLABEL		0xFFFF     /* max. of unsigned short */
#define LABELMAX        NOLABEL
#ifdef RECT
#define NORECT          0xFFFF     /* max. of Coordinate */
#define NOCB            NOLABEL
#endif /* RECT */

#define YES		1
#define NO		0
#define RIGHTANGLE	90
#define M_PI		3.14159265358979323846

/* for sweep2 */
#define LE		0
#define RE		1
#define DELETED		-2

/* An end point of a Voronoi edge is on the border of an image,
   its sitenbr is FRAME (erase.c, output.c) */
#define	FRAME		-1

/* for hash table */
#define M1		10007	/* prime number */
#define M2		3	/* prime number */
#define NIL		0
#define NODATA		-1

/* default values */
#ifndef RECT
#define SAMPLE_RATE	7
#define	NOISE_MAX	20
#define FREQ_RATE	0.5
#define Ta_CONST        40
#define SMWIND          2    /* for smoothing the freq. distribution
			        of distance */
#else /* for Manhattan layout --- just an example */
#define SAMPLE_RATE	7
#define	NOISE_MAX	20
#define FREQ_RATE	0.03
#define Ta_CONST        200
#define SMWIND          2    /* for smoothing the freq. distribution
			        of distance */
#endif /* RECT */


#ifdef RECT
/* for connected components */  
#define CRECT_WIDTH_MAX    500
#define CRECT_HEIGHT_MAX   500
/* #define CRECT_AREA_MAX     5000 */
#define CRECT_ASPECT_MAX   5.0 

/* for blocks Song */
#define BRECT_ASPECT_MAX   4.0
#define BRECT_VAREA_MIN    40000
#define BRECT_AREA_MIN     50
#endif /* RECT */
                          
/* ܥΥդνϾ */
#define OUTPUT		1
#define NO_OUTPUT	0

#ifdef RECT
#define OUT             1
#define IN              0
#endif /* RECT */

typedef unsigned short Coordinate;
typedef unsigned short Label;
typedef unsigned int   NumPixel;
typedef unsigned int   Key;
typedef unsigned int   HashVal;
typedef double         Coord;

/* ٥ȥɽ¤
   the structure for a vector */
typedef struct{
  int x;
  int y;
} Vector;

/* ɽ¤
   the structure for a binary image */ 
typedef struct{
  char *image;
  Coordinate imax, jmax; /* width and height of the image */
} ImageData;

/* ǤΥ٥Ⱥɸɽ¤
   the structure for a pixel */
typedef struct{
  Label label;     /* label */
  Coordinate xax,yax; /* coordinates */
} BlackPixel;

/* Ϣʬνſɸ, ϤνĲĹɽ¤ */
/*
typedef struct{
    unsigned short x,y;
    unsigned short dx,dy;
    unsigned int bpn;
} Component;
*/

/* ܤϢʬδطɽ¤
   the structure for a neighboring relation
   between connected components(CC's)

         lab2
       -------
       |  x  |
       -------
    --- /
    | |/ angle
    |x|-----
    | |
    ---
    lab1
*/
typedef struct{
  float dist;       /* min. distance between CC's */
  float angle;      /* angle between CC's */
  Label lab1, lab2; /* labels of CC's */
} Neighborhood;

/* ܥΥդλ, ɤϢʬ֤Ǥ뤫ɽ¤
   the structure for a Voronoi edge */
typedef struct{
  int	sp,ep;
  Coordinate xs,xe,ys,ye;  /* + (xs,ys)
				\
				\  Voronoi edge
				\
				+ (xe,ye)
				*/
  Label lab1,lab2;  /* this Voronoi edge is between
			      CC of a label "lab1" and that of lab2 */
  unsigned short yn;
} LineSegment;

/* ܥΥդü
   the structure for a Voronoi point */
typedef struct node{
  int line;
  struct node *next;
} EndPoint;

/* ϢʬΥ٥Фϥåɽ
   the structure for a hash table for
   representing labels of CC's */
typedef struct hash {
  Label lab1;
  Label lab2;
  unsigned int entry;
  struct hash *next;
} HashTable;

/* ɽ¤
   the structure for a rectangle */
typedef struct{
  Coordinate xs,xe,ys,ye;
} Rect;

struct Freenode {
    struct Freenode	*nextfree;
};

struct Freelist {
    struct Freenode	*head;
    int			nodesize;
};

struct Point {
    float		x,y;
};

/* structure used both for sites and for vertices */
struct Site {
    struct Point	coord;
    int			sitenbr;
    int			refcnt;
    unsigned int	label;
};

struct Edge	{
    float		a,b,c;
    struct	Site 	*ep[2];
    struct	Site	*reg[2];
    int			edgenbr;
    unsigned int	lab1,lab2;
    float		dist;
};

struct Halfedge {
    struct Halfedge	*ELleft, *ELright;
    struct Edge		*ELedge;
    int			ELrefcnt;
    char		ELpm;
    struct Site		*vertex;
    float		ystar;
    struct Halfedge	*PQnext;
};

#endif /* DEFS_H_INCLUDED_ */

/*
   $Id: pset.h,v 1.1 2000/09/19 22:20:12 maosong Exp maosong $
   $Date: 2000/09/19 22:20:12 $
   $Revision: 1.1 $
   $Author: maosong $
*/
#ifndef FUNCTION_H_INCLUDED_
#define FUNCTION_H_INCLUDED_
#include <stdio.h>
#include <stdlib.h>

/*
   $Id: pset.h,v 1.1 2000/09/19 22:20:12 maosong Exp maosong $
   $Date: 2000/09/19 22:20:12 $
   $Revision: 1.1 $
   $Author: maosong $
*/

#ifndef READ_IMAGE_INCLUDED_
#define READ_IMAGE_INCLUDED_
#ifndef O_BINARY
#define O_BINARY        0
#endif

#define II_TIFF_VER    0x2a00
#define RAS_MAGIC      0x59a66a95
#define II_RAS_MAGIC   0x956aa659
#define MULTIPLY       16 /* the image width of sunraster 
			     must be a multiply of 16 */
/*
   the header of sun rasterfiles.
   see "man rasterfile"
*/
struct rasterfile{
  int  ras_magic;
  int  ras_width;
  int  ras_height;
  int  ras_depth;
  int  ras_length;
  int  ras_type;
  int  ras_maptype;
  int  ras_maplength;
};

#endif



/* bit_func.c */
int bit_get( ImageData *, Coordinate, Coordinate );
void bit_set( ImageData *, Coordinate, Coordinate, int );
int byte_pos ( char, int );
void frame( ImageData *, int, int );

/* cline.c */
void analyze_cline( char **, int *, int * );

/* dinfo.c */
void dparam();
void dnumber( int, int );
#ifdef TIME
void dtime();
#endif

/* edgelist.c */
void ELinitialize();
struct Halfedge *HEcreate( struct Edge *, int );
void ELinsert( struct Halfedge *, struct Halfedge * );
struct Halfedge *ELgethash( int );
struct Halfedge *ELleftbnd( struct Point * );
void ELdelete( struct Halfedge * );
struct Halfedge *ELright( struct Halfedge * );
struct Halfedge *ELleft( struct Halfedge * );
struct Site *leftreg( struct Halfedge * );
struct Site *rightreg( struct Halfedge * );

/* erase.c */
int start_pos ( int );
int end_pos ( int );
unsigned int Dh_ave( unsigned int *, int );
void hist();
int distinction( Label, Label, int );
void erase_endp( int );
void erase_aux();
void erase();
void init_u_int( unsigned int * );
void init_int( int * );

/* geometry.c */
void geominit();
struct Edge *bisect( struct Site *, struct Site * );
struct Site *intersect( struct Halfedge *, struct Halfedge * );
int right_of( struct Halfedge *, struct Point * );
void endpoint( struct Edge *, int, struct Site *, Coordinate, Coordinate );
float dist( struct Site *, struct Site * );
void makevertex( struct Site * );
void deref( struct Site * );
void ref( struct Site * );

/* hash.c */
HashVal hash1( Key );
HashVal hash2( Key );
void init_hash();
Key key( Label, Label );
int search( Label, Label );
void enter( Label, Label, unsigned int );

/* heap.c */
void PQinsert( struct Halfedge *, struct Site *, float );
void PQdelete( struct Halfedge * );
int PQbucket( struct Halfedge * );
int PQempty();
struct Point PQ_min();
struct Halfedge *PQextractmin();
void PQinitialize();

/* img_to_site.c */
void img_to_site( ImageData * );
#ifdef RECT
int  is_noise_cc( int, char, Rect );
void bf_edgelab_smpl( ImageData *, Coordinate, Coordinate, Label, char);
#else
int  is_noise_cc( int );
void bf_edgelab_smpl( ImageData *, Coordinate, Coordinate, Label);
#endif /* RECT */
Vector next_point( ImageData *, Coordinate *, Coordinate *,
		   Vector *, Vector, Label );
Vector first_d( Vector );
Vector rot_d( Vector );
void edge_lab( Vector, Vector, Coordinate, Coordinate, Label );
void bpxset( Coordinate, Coordinate, Label );
int scomp( const void *, const void * );

/* label_func.c */
void lab_format( ImageData * );
Label lab_get( Coordinate, Coordinate );
void lab_set( Coordinate, Coordinate, Label );
void free_limg();

/* memory.c */
void freeinit( struct Freelist *, int );
char *getfree( struct Freelist * );
void makefree( struct Freenode *, struct Freelist * );
char *myalloc( unsigned );
char *myrealloc( void *, unsigned, unsigned int, size_t);

/* output.c */
void in_frame( float *, float *, float, struct Edge *, int,
	       Coordinate, Coordinate );
void s_in_frame( float *, float *, float *, float *,
		 struct Edge *, Coordinate );
void e_in_frame( float *, float *, float *, float *,
		 struct Edge *, Coordinate, Coordinate );
void frameout( float *, float *, float *, float *,
	       int *, int *, struct Edge *, Coordinate, Coordinate );
void out_ep2( struct Edge *, struct Site *, Coordinate, Coordinate );


/* read_image.c */
void read_image( char *, ImageData * );
int tiff2imgd( char *, ImageData * );
int tiff_p( char * );
int ras2imgd( char *, ImageData * );
void swab_rashead( struct rasterfile * );

/* sites.c */
struct Site *nextsite();

/* usage.c */
void usage();

/* voronoi.c */
void voronoi( Coordinate, Coordinate );

/* brect.c */
#ifdef RECT
int  is_noise_crect( Rect );
int  rwidth( Rect );
int  rheight( Rect );
int  rarea( Rect );
int  raspect( Rect );
void init_crect();
void resize_crect();
void crectset( Rect, Label );
void generate_brect();
void init_brect();
void init_cb_tab();
void resize_brect();
void set_brect( Label, Label );
Rect merge_rect( Rect, Rect );
int  is_noise_brect( Rect );
#endif /* RECT */
#endif /* FUNCTION_H_INCLUDED_ */

/*
   $Id: pset.h,v 1.1 2000/09/19 22:20:12 maosong Exp maosong $
   $Date: 2000/09/19 22:20:12 $
   $Revision: 1.1 $
   $Author: maosong $
*/

#include<stdio.h>

extern float	xmin, xmax, ymin, ymax, deltax, deltay;

extern struct	Site		*sites;
extern int			nsites;
extern int			siteidx;
extern int			sqrt_nsites;
extern int			nvertices;
extern struct 	Freelist 	sfl;
extern struct	Site		*bottomsite;

extern int 			nedges;
extern struct	Freelist 	efl;

extern struct   Freelist	hfl;
extern struct	Halfedge	*ELleftend, *ELrightend;
extern int 			ELhashsize;
extern struct	Halfedge	**ELhash;

extern int 			PQhashsize;
extern struct	Halfedge 	*PQhash;
extern int 			PQcount;
extern int 			PQmin;

/* extern Component                *component; */
extern BlackPixel               *bpx;
extern Neighborhood	        *neighbor;
extern LineSegment	        *lineseg;
extern EndPoint		        *endp;
extern HashTable	        *hashtable[M1+M2];

extern NumPixel                 BPnbr;
extern Label                    LABELnbr;
extern unsigned int	        NEIGHnbr;
extern unsigned int	        LINEnbr;
extern unsigned int             sample_pix;
extern unsigned int             point_edge;
extern unsigned int             edge_nbr;
extern long		        SiteMax;

extern int		        sample_rate;
extern int		        noise_max;
extern float                    freq_rate;
extern int                      smwind;
extern int                      Ta;

extern char                     output_pvor;
extern char                     output_avor;
extern char                     display_parameters;

#ifdef TIME
extern float    b_s_time;
extern float    r_time;
extern float    v_time;
extern float    e_time;
extern float    o_time;
#endif /* TIME */

#ifdef RECT
extern int     crect_width_max;
extern int     crect_height_max;
/* extern int  crect_area_max; */
extern float   crect_aspect_max;

extern float   brect_aspect_max;
extern float   brect_varea_min;
extern float   brect_area_min;

extern Rect    *crect;
extern Rect    *brect;
extern Label   *cb_tab;
extern Label   BRECTid;
extern Label   BRECTnbr;
#endif /* RECT */
/*
**  File: util.h
**  Purpose: basic c utilities to get started coding with: new, fOpen, fClose, err.
**  	Each of these is described below.  You will have to create your own
**  	Usage() function!
**  Author: Gregory Marton (gremio@cfar.umd.edu)
**  Organization: Language and Media Processing Lab
**                 Center for Automation Research
**                   University of Maryland
**                   College Park, MD 20742
**  $Id: pset.h,v 1.1 2000/09/19 22:20:12 maosong Exp maosong $
**
*/

#ifndef UTIL_H
#define UTIL_H

/*static char *UTIL_H_VERSION="$Id: pset.h,v 1.1 2000/09/19 22:20:12 maosong Exp maosong $"; */
/* Utilities:  Gregory Marton
 * basic c utilities to get started coding with: new, fOpen, fClose, err.
 * Each of these is described below.  You will have to create your own 
 * Usage() function! 
 * 
 * Filename: util.h, util.c
 * Version: $Id: pset.h,v 1.1 2000/09/19 22:20:12 maosong Exp maosong $
 * Usage:   #include "util.h";  
 *          then define a Usage() function
 * Assumptions: maximum file name length is 256 characters.
 *              mode for fOpen is specified according to K&R ANSI standard 
 *                for fopen().
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <assert.h>

#define NDBG

/* static char *progname; */ /* in first line of main, "progname=argv[0];" */

#define new(type, nobj) ((type *) Alloc( (size_t) nobj * sizeof(type)))
void *Alloc(size_t Size);
/* allocates space for number new types and returns pointer to the array.
 * new(int, 1) produces a single new integer. 
 * usage:  variable=new(type_of_objects, number_of_objects);  */

FILE *fOpen(const char *filename, const char *mode);
/* if the filename is '-' or is empty, returns stdin/stdout depending on 
 *   the mode.
 * otherwise if the file named exists, opens the file for the mode, and
 *   returns the file pointer. 
 * if that file does not exist, generates a USAGE error, and returns NULL.
 */


void fClose(FILE *f);

#define WARN 1
#define CONT 5
#define EXIT 10
#define USAGE 15

void err(char *message, int contin);

void Usage();
/* Please define this function somewhere.  It is not defined in util.c but
 * must be present for fOpen to work correctly.  It should print out a 
 * simple usage message.  Something of the form
 * printf("usage: %s [-o outputfile] <file1> <file2>\n", progname);  */

int sh_c(const char *command, 
	 char *standardout, const int standardoutmax, 
	 char *standarderr, const int standarderrmax);
  /* this is intended to act like `sh -c command >standardout 2> standarderr` 

     sh will be asked to perform action.  
     at most standardoutmax characters of standard output will be assigned
        to the string standardout, which must be able to contain them.
     at most standarderrormax characters of standard error output from the 
        command will similarly be assigned to standarderror, which also must 
	be previously allocated.
     returns the exit status of the command.
     
     NOTE: as a general rule, this function is slow.
     NOTE: check this with 'assert(!sh_c(...));' as most functions, returning 
     success, 0, will raise an assertion failed. 
     */

#define chop(s);  (s)[strlen(s)-1]=0;


#endif

