/*  
**  File: Simplex.c
**  Purpose: optimize a practical objective function (where there is no
**    explicit mathematical form) using Nelder_Mead's simplex algorithm
**  Modified by: Song Mao (maosong@cfar.umd.edu)
**  Date:   22 March 1999
**  Organization: Language and Media Processing Lab
**                 Center for Automation Research
**                   University of Maryland
**                   College Park, MD 20742
**  References: Numerical Recipes in C
**  $Id: Simplex.c,v 1.1 2000/09/19 22:16:27 maosong Exp $
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "pset.h"

int Simplex(TRAIN_PROTOCOL *train_protocol, SIMPLEX_PARAM *opt_param,
	    void *ben_param, void *seg_param, char *weight_file,
	    void (*genscore)(TRAIN_PROTOCOL *, void *, void *, char *, float *, SCORE **),
	    char **train_report_p, char **opt_seg_param_p)
{
	FILE  *fp, *fpopt;
   	int i, j, ilo, ihi, inhi, mpts = opt_param->ndim + 1;
   	float *psum;
   	float *pbar;
   	float ybar, tbar;
   	float ytry, ysave, sum, rtol;
   	int nFuncEval;
	float *plow;
	float ylow;
	SCORE *score;

	score = (SCORE *) malloc (sizeof(SCORE));

	fp = fopen((*train_report_p), "a");
	fpopt = fopen((*opt_seg_param_p), "w");

        fprintf(fp,"# Feval\t");
	for(i = 1; i <= opt_param->ndim; i++)
		fprintf(fp,"p[%d]\t",i);
	fprintf(fp,"score\ttiming\t");
	for(i = 1; i <= opt_param->ndim; i++)
                fprintf(fp,"plow[%d]\t",i);
	fprintf(fp,"Flow\n");

        for(i = 2; i<= opt_param->ndim + 1; i++) {		/* initialize simplex */
          for(j = 1; j<= opt_param->ndim; j++) {
            if(j == i-1)
              opt_param->p[i][j] = opt_param->p[1][j] + opt_param->scale[j];
            else
              opt_param->p[i][j] = opt_param->p[1][j];
          }
        }

	nFuncEval = 0;
	ylow = 1.0;
   	plow = (float *) malloc ( (opt_param->ndim + 1) * sizeof(float) );
        for(i = 1; i <= opt_param->ndim + 1; i++) {
	  (*genscore)(train_protocol, seg_param, ben_param, weight_file, opt_param->p[i], &score);
	  nFuncEval ++;
          opt_param->y[i] = score->err_rate;
          opt_param->t[i] = score->timing;

	  if(ylow > opt_param->y[i]) {			/* print result in report file */
		ylow = opt_param->y[i];
		for(j = 1; j <= opt_param->ndim; j++)
			plow[j] = opt_param->p[i][j];
	  }
	  fprintf(fp,"%d\t", nFuncEval);
	  for(j = 1; j <= opt_param->ndim; j++)
		fprintf(fp,"%.3f\t", opt_param->p[i][j]);
	  fprintf(fp,"%.3f\t", opt_param->y[i] * 100);
	  fprintf(fp,"%.1f\t", opt_param->t[i]);
          for(j = 1; j <= opt_param->ndim; j++)
                fprintf(fp,"%.3f\t", plow[j]);
	  fprintf(fp,"%.3f\n", ylow * 100);
	}
	free(plow);

   	psum = (float *) malloc ( (opt_param->ndim + 1) * sizeof(float) );
   	pbar = (float *) malloc ( (opt_param->ndim + 1) * sizeof(float) );

   	GET_PSUM

   	for(;;) {
     	  ilo = 1;
     	  ihi = opt_param->y[1] > opt_param->y[2] ? (inhi = 2,1) : (inhi = 1,2);
     	  for (i = 1; i <= mpts; i++) {
            if(opt_param->y[i] < opt_param->y[ilo]) ilo = i;
       	    if(opt_param->y[i] > opt_param->y[ihi]) {
              inhi = ihi;
              ihi = i;
            } else if (opt_param->y[i] > opt_param->y[inhi])
            if (i != ihi) inhi = i;
      	  }

  	  fflush(NULL);
	  if(opt_param->criflg == NR)		 /* Numerical recipe stop criterion */
     		  rtol = 2.0 * fabs(opt_param->y[ihi] - opt_param->y[ilo])/
			(fabs(opt_param->y[ihi]) + fabs(opt_param->y[ilo])); 
	  else if(opt_param->criflg == NM) {   	/* Nelder-Mead stop criterion */
     		  rtol = 0.0;
     		  for (j=1; j <= opt_param->ndim; j++) 
		    pbar[j] = (psum[j] - opt_param->p[ihi][j])/opt_param->ndim;
		  (*genscore)(train_protocol, seg_param, ben_param, weight_file, pbar, &score);
     		  nFuncEval++;
     		  ybar = score->err_rate;
		  tbar = score->timing;

                  fprintf(fp,"%d\t", nFuncEval);    /* print result in report file */
                  for(j = 1; j <= opt_param->ndim; j++)
                    fprintf(fp,"%.3f\t", pbar[j]);
                  fprintf(fp,"%.3f\t", ybar * 100);
                  fprintf(fp,"%.1f\t", tbar);
                  for(j = 1; j <= opt_param->ndim; j++)
                    fprintf(fp,"%.3f\t",  opt_param->p[ilo][j]);  /* print plow and ylow */
                  fprintf(fp,"%.3f\n", opt_param->y[ilo] * 100);

     		  for (i=1; i <= mpts; i++) 
		    rtol += (opt_param->y[i] - ybar) * (opt_param->y[i] - ybar);
     		  rtol = rtol / opt_param->ndim;
     		  rtol = sqrt(rtol);
	  }
	  else {
		  printf("No such stop criterion.\n");
		  exit(-1);
	  }

     	  if(rtol < opt_param->ftol) 
       		  break;
     	  if(nFuncEval >= opt_param->nmax) { 
       		  printf("TOO many iterations in downhill_simplex\n");
       		  break;
     	  }

	  /* simplex reflection */
     	  ytry = Amotry(train_protocol, seg_param, ben_param, weight_file, opt_param, psum,
			genscore, ihi, &nFuncEval, -(opt_param->alpha), fp); 
          for(j = 1; j <= opt_param->ndim; j++)
            fprintf(fp,"%.3f\t", opt_param->p[ilo][j]);  /* print plow and ylow */
          fprintf(fp,"%.3f\n", opt_param->y[ilo] * 100);

     	  if(ytry <= opt_param->y[ilo]) {

		  /* simplex expension */
        	  ytry = Amotry(train_protocol, seg_param, ben_param, weight_file, opt_param,
			psum, genscore, ihi, &nFuncEval, opt_param->gamma, fp);
          	  for(j = 1; j <= opt_param->ndim; j++)
                    fprintf(fp,"%.3f\t", opt_param->p[ilo][j]);  /* print plow and ylow */
          	  fprintf(fp,"%.3f\n", opt_param->y[ilo] * 100);
	  }
     	  else if (ytry >= opt_param->y[inhi]) {
       		  ysave = opt_param->y[ihi];

		  /* simplex contraction */
       		  ytry = Amotry(train_protocol, seg_param, ben_param, weight_file, opt_param, psum,
			genscore, ihi, &nFuncEval, opt_param->beta, fp);
 	          for(j = 1; j <= opt_param->ndim; j++)
         	    fprintf(fp,"%.3f\t", opt_param->p[ilo][j]); /* print plow and ylow */
 	          fprintf(fp,"%.3f\n", opt_param->y[ilo] * 100);

       		  if(ytry >= ysave) {
         	    for(i = 1; i<=mpts; i++) {
           		  if(i != ilo) {
             			  for(j = 1; j <= opt_param->ndim; j++) {
					 
					  /* simplex shrinkage */
               				  psum[j] = opt_param->sigma * (opt_param->p[i][j] + opt_param->p[ilo][j]);
               				  opt_param->p[i][j] = psum[j];
             			  }
				  (*genscore)(train_protocol, seg_param, ben_param, weight_file, psum, &score);
				  nFuncEval++;
             			  opt_param->y[i] = score->err_rate;
             			  opt_param->t[i] = score->timing;
 
 		                  fprintf(fp,"%d\t", nFuncEval);	/* print result in report file */
                 		  for(j = 1; j <= opt_param->ndim; j++)
   		                    fprintf(fp,"%.3f\t", psum[j]);
                  		  fprintf(fp,"%.3f\t", opt_param->y[i] * 100);
 		                  fprintf(fp,"%.1f\t", opt_param->t[i]);
                 		  for(j = 1; j <= opt_param->ndim; j++)
                    		    fprintf(fp,"%.3f\t", opt_param->p[ilo][j]);
 		                  fprintf(fp,"%.3f\n", opt_param->y[ilo] * 100);

           		  }
         	    }
         	    GET_PSUM;
       		  }
     	  }
     }
     fprintf(fp, "\nOptimal_Parameter_Vector = ");
     for(j = 1; j <= opt_param->ndim; j++)
     	fprintf(fp,"%.3f\t", opt_param->p[ilo][j]);
     fprintf(fp, "\n");
     fprintf(fp, "Optimal_Performance_Value = ");
     fprintf(fp,"%.3f\n", opt_param->y[ilo] * 100);

     switch(train_protocol->seg_alg) {
       case XYCUT:
	 fprintf(fpopt, "# Optimal parameter values for ");
	 fprintf(fpopt, "the X-Y cut page segmentation algorithm.\n");
	 if(((XYCUT_PARAM *)(seg_param))->alg_mode == FUNC_CALL) 
	   fprintf(fpopt, "ALG_MODE	=	func_call\n");
	 else
	   fprintf(fpopt, "ALG_MODE      =       shell_call\n");
	 fprintf(fpopt, "TNX	=	%d\n", (int)opt_param->p[ilo][1]);
	 fprintf(fpopt, "TNY	=	%d\n", (int)opt_param->p[ilo][2]);
	 fprintf(fpopt, "TCX	=	%d\n", (int)opt_param->p[ilo][3]);
	 fprintf(fpopt, "TCY	=	%d\n", (int)opt_param->p[ilo][4]);
         fclose(fpopt);
	 break;
       case DOCSTRUM:
	 fprintf(fpopt, "# Optimal parameter values for ");
	 fprintf(fpopt,"the Docstrum page segmentation algorithm.\n");
	 if(((DOCSTRUM_PARAM *)(seg_param))->alg_mode == FUNC_CALL) 
	   fprintf(fpopt, "ALG_MODE	=	func_call\n");
	 else
	   fprintf(fpopt, "ALG_MODE      =       shell_call\n");
	 fprintf(fpopt, "CONCOM_DIR	=	%s\n", ((DOCSTRUM_PARAM *)(seg_param))->concom_dir);
	 fprintf(fpopt, "CONCOM_SUFFIX	=	%s\n", ((DOCSTRUM_PARAM *)(seg_param))->concom_suffix);
	 fprintf(fpopt, "KNN		=	%d\n", ((DOCSTRUM_PARAM *)(seg_param))->knn);
	 fprintf(fpopt, "LOWSIZE        =       %d\n", ((DOCSTRUM_PARAM *)(seg_param))->lsize);
	 fprintf(fpopt, "HIGHSIZE        =      %d\n", ((DOCSTRUM_PARAM *)(seg_param))->hsize);
	 fprintf(fpopt, "HANGTOL         =       %d\n", ((DOCSTRUM_PARAM *)(seg_param))->hatol);
	 fprintf(fpopt, "VANGTOL         =       %d\n", ((DOCSTRUM_PARAM *)(seg_param))->vatol);
	 fprintf(fpopt, "TCC             =       %f\n", opt_param->p[ilo][1]);
	 fprintf(fpopt, "PAR             =       %f\n", opt_param->p[ilo][2]);
	 fprintf(fpopt, "PER             =       %f\n", opt_param->p[ilo][3]);
	 fprintf(fpopt, "SCR             =       %f\n", ((DOCSTRUM_PARAM *)(seg_param))->scr);
	 fprintf(fpopt, "DIS             =       %d\n", (int)opt_param->p[ilo][4]);
         fclose(fpopt);
	 break;
       case VORONOI:
	 fprintf(fpopt, "# Optimal parameter values for the ");
	 fprintf(fpopt, "Voronoi-based page segmentation algorithm.\n");
         if(((VORONOI_PARAM *)(seg_param))->alg_mode == FUNC_CALL) 
           fprintf(fpopt, "ALG_MODE     =       func_call\n");
         else
           fprintf(fpopt, "ALG_MODE      =       shell_call\n");
         fprintf(fpopt, "SR    =       %d\n", (int)opt_param->p[ilo][1]);
         fprintf(fpopt, "NM    =       %d\n", (int)opt_param->p[ilo][2]);
         fprintf(fpopt, "FR    =       %f\n", opt_param->p[ilo][3]);
         fprintf(fpopt, "TA    =       %d\n", (int)opt_param->p[ilo][4]);
         fclose(fpopt);        
         break;
       default:
	 printf("No such page segmentation algorithm in the PSET package.\n");
	 exit(-1); 
     }

     fprintf(fp, "\n# End of the training.\n");
     free(psum); 
     free(pbar);
     free(score);
     fclose(fp);

     return(0);
}
