/*  
**  File: XycutFunction.c
**  Purpose: X-Ycut segmentation algorithm function implementation
**  Author: Song Mao (maosong@cfar.umd.edu)
**  Date:   18 January 1999
**  Organization: Language and Media Processing Lab
**                 Center for Automation Research
**                   University of Maryland
**                   College Park, MD 20742
**  References:
**
**  $Id: XycutFunction.c,v 1.1 2000/09/19 22:16:27 maosong Exp $
*/

#include "rafport.h"
#include <sys/types.h>
#include <sys/uio.h>
#include "dafstype.h"
#include "rlimage.h"
#include "dafs.h"

#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h" 
#include "tiff.h"
#include <time.h>

#ifdef SUNOS
#include "libfun.h"
#endif

#include "pset.h"

int XycutFunction(char *tif_filename, XYCUT_PARAM *para, char *dafs_filename)
{
	char **buf;
	int **x_hist_table, **y_hist_table;

	DOCUMENT_IMAGE *document;
	NODE_XYCUT *root;


	/*------------------------------------------------------------------
  		Read Tiff Pixel Matrix into a buffer - buf
	-------------------------------------------------------------------*/
	ReadTIFFImageXycut(tif_filename, &para, &document, &root, &buf);


	/*------------------------------------------------------------------
  		Create X and Y histogram look-up table
	-------------------------------------------------------------------*/
	Create_Histogram_Table(buf, document, &x_hist_table, &y_hist_table); 


	/*------------------------------------------------------------------
  		Make X-Y-Cuts tree structure  
	-------------------------------------------------------------------*/
	MakeTree(tif_filename, buf, root, document, x_hist_table, y_hist_table, para, dafs_filename);

	/*------------------------------------------------------------------
  		Free Jobs  
	-------------------------------------------------------------------*/
	FreeJobXycut(document, para, buf, x_hist_table, y_hist_table);
	
	return(0);

}/*main*/



/* < READ TIFF IMAGE */
int ReadTIFFImageXycut( char *infile, XYCUT_PARAM **para_p, DOCUMENT_IMAGE **document_p,
			NODE_XYCUT **root_p, char ***buf_p)
{
  	char * tmp_buf;
  	int i, j, k;
  	TIFF *tif;
  	long row;

  	int ii;
  	int  photoint;
  	short shorttag;
  	float floattag;

  	*root_p=(NODE_XYCUT *) malloc (sizeof (NODE_XYCUT));
  	*document_p=(DOCUMENT_IMAGE *) malloc (sizeof (DOCUMENT_IMAGE));

  	tif = TIFFOpen(infile, "r");

  	if (tif == NULL) 
     		return(-1);

  	TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &(*document_p)->width);

  	TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &(*document_p)->height);

  	TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &shorttag);
  	photoint = (int)shorttag;

  	TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &shorttag);
  	(*document_p)->depth = (int)shorttag;

  	floattag = 0;
  	TIFFGetField(tif, TIFFTAG_XRESOLUTION, &floattag);
  	(*document_p)->xres = (int) floattag;

  	floattag = 0;
  	TIFFGetField(tif, TIFFTAG_YRESOLUTION, &floattag);
  	(*document_p)->yres = (int) floattag;

        (*para_p)->Tny = (*para_p)->Tny * (*document_p)->xres / 300;
        (*para_p)->Tnx = (*para_p)->Tnx * (*document_p)->yres / 300;
        (*para_p)->Tcx = (*para_p)->Tcx * (*document_p)->xres / 300;
        (*para_p)->Tcy = (*para_p)->Tcy * (*document_p)->yres / 300;

  	(*document_p)->image = (char *) malloc ( (*document_p)->height *
					TIFFScanlineSize(tif)+10);
  	tmp_buf = (*document_p)->image;

  	if ( (*document_p)->depth == 8)
      	(*document_p)->rowsize = (*document_p)->width;
  	else  if ( (*document_p)->depth == 1)
    	(*document_p)->rowsize = ( (*document_p)->width+7)>>3;

  	for (row = 0; row < (*document_p)->height; row++) {
    	TIFFReadScanline(tif, &tmp_buf[row* (*document_p)->rowsize], row, 0);
  	}

  	if (photoint == PHOTOMETRIC_MINISBLACK) {
    		for (ii=0;ii< (*document_p)->height * (*document_p)->rowsize;ii++)
	 	tmp_buf[ii] = 255-tmp_buf[ii]; 
     	}

  	(*root_p)->info="root";
  	(*root_p)->ul_x=0;
  	(*root_p)->ul_y=0;
  	(*root_p)->width=(*document_p)->width;
  	(*root_p)->height=(*document_p)->height;
  	(*root_p)->local_Tnx = (*para_p)->Tnx;
  	(*root_p)->local_Tny = (*para_p)->Tny;

  	*buf_p=(char **) malloc ( (*document_p)->height * sizeof(char *));
  	for(i=0; i<(*document_p)->height; i++)
    	(*buf_p)[i]=(char *) malloc ((*document_p)->width * sizeof(char));

  	for(i=0; i< (*document_p)->height; i++) 
    	for(j=0; j < (*document_p)->rowsize; j++) 
      	for(k = 8; k > 0; k--)
      		(*buf_p)[i][j*8+8-k] = (tmp_buf[i * (*document_p)->rowsize+j]>>(k-1)) & 1;

  	TIFFClose(tif);
  	return(0);
}

/* > */

/* < Create X and Y histogram look-up table */ 
void Create_Histogram_Table(char **buf, DOCUMENT_IMAGE *document,
			   int ***x_hist_table_p, int ***y_hist_table_p)
{
  	int i,j;
  	(*x_hist_table_p)=(int **) malloc (document->height * sizeof(int*));
  	(*y_hist_table_p)=(int **) malloc (document->height * sizeof(int*));

  	for(i=0; i<document->height; i++) {
    		(*x_hist_table_p)[i]=(int *) malloc (document->width * sizeof(int));
    		(*y_hist_table_p)[i]=(int *) malloc (document->width * sizeof(int));
  	}

  	for(i=0; i<document->height ; i++) {
    		for(j=0; j<document->width; j++) {
      			if(j==0) 
        			(*y_hist_table_p)[i][j]=buf[i][j];
      			else
        			(*y_hist_table_p)[i][j]=buf[i][j] +
				(*y_hist_table_p)[i][j-1];
    		}
  	}

  	for(i=0; i<document->height ; i++) {
    		for(j=0; j<document->width; j++) {
      			if(i==0) 
        			(*x_hist_table_p)[i][j]=buf[i][j];
      			else
        			(*x_hist_table_p)[i][j]=buf[i][j] +
				(*x_hist_table_p)[i-1][j];
    		}
  	}
  	
}

/* > */

/* < Zone Adjusting image, i.e. get text region of a Tiff Image */ 
void ZoneAdjust(char **buf, NODE_XYCUT *root, DOCUMENT_IMAGE *document,
		int **x_hist_table, int **y_hist_table, NODE_XYCUT **zone_p,
		int ***free_hist_p, int *h_p)
{
	int i=0, j=0, k=0, l=0;

	(*zone_p)->local_Tnx = ((*zone_p)->height) * root->local_Tnx/root->height;
	(*zone_p)->local_Tny = ((*zone_p)->width) * root->local_Tny/root->width;

	GetHistogram(zone_p, x_hist_table, y_hist_table, document); 

	(*free_hist_p)[(*h_p)++] = (*zone_p)->x_histogram;
	(*free_hist_p)[(*h_p)++] = (*zone_p)->y_histogram;

	/* find the text region, i.e. ignore blank margins of image */

	while( (*zone_p)->x_histogram[i] <= (*zone_p)->local_Tnx && i <
		(*zone_p)->width - 1 ) 
  		i++;
		(*zone_p)->ul_x=(*zone_p)->ul_x+i;

	j=(*zone_p)->width - 1;
	while((*zone_p)->x_histogram[j] <= (*zone_p)->local_Tnx && j > i ) 
	j--;
	(*zone_p)->width=j+1-i;

	k=0;
	while((*zone_p)->y_histogram[k] <= (*zone_p)->local_Tny && k <
						(*zone_p)->height - 1 )
	k++;
	(*zone_p)->ul_y=(*zone_p)->ul_y+k;

	l = (*zone_p)->height - 1;
	while((*zone_p)->y_histogram[l] <= (*zone_p)->local_Tny && l > k) 
	l--;
	(*zone_p)->height = l+1-k;

	(*zone_p)->local_Tnx = ((*zone_p)->height) * root->local_Tnx/root->height;
	(*zone_p)->local_Tny = ((*zone_p)->width) * root->local_Tny/root->width;

	GetHistogram(zone_p, x_hist_table, y_hist_table, document); 

	(*free_hist_p)[(*h_p)++] = (*zone_p)->x_histogram;
	(*free_hist_p)[(*h_p)++] = (*zone_p)->y_histogram;

}

/* > */

/* < Make X-Y-Cuts tree structure  subroutine */
int  MakeTree(char *tif_filename, char **buf, NODE_XYCUT  *root,
		DOCUMENT_IMAGE *document, int **x_hist_table,
		int **y_hist_table, XYCUT_PARAM *para, char *dafs_filename)
{
	int i;
	/* dafs */
	iEntityPtr dafs_doc;
	iEntityPtr dafs_zone;
	iDAFSError dafs_err;
	iBox dafs_box;
	iImagePtr imPtr;
	/* push and pop */
	int sp=0;
	int leaf_sp=0;
	NODE_XYCUT **stack, **leaf_stack;
	NODE_XYCUT *end_node;

	int updated_leftwidth, updated_upperheight;
	int  cut_direction;
	NODE_XYCUT *x, *y, *t;

	NODE_XYCUT **free_zone;
	int  l = 0;
	int  **free_hist;
	int  h = 0;

	NODE_XYCUT *z;

	z = (NODE_XYCUT *) malloc (sizeof (NODE_XYCUT));
	z->l = z;
	z->r = z;

	/* allocate stack and leaf_stack */
	stack=(NODE_XYCUT **) malloc (MAX * sizeof(NODE_XYCUT *));
	leaf_stack=(NODE_XYCUT **) malloc (MAX * sizeof(NODE_XYCUT *));

	free_zone = (NODE_XYCUT **) malloc (MAX * sizeof(NODE_XYCUT *));
	free_hist = (int **) malloc (MAX * sizeof(int *));

	/* set leaf to z node */
	PushLeaf(z, &leaf_sp, &leaf_stack);

	/* Dafs file initialization */
	dafs_err=i_InitLib();
	dafs_doc=i_NewEntity(iDocumentT, 0);
	dafs_box.x=0;
	dafs_box.y=0;
	dafs_box.w=document->width;
	dafs_box.h=document->height;
	i_SetBox(dafs_doc,dafs_box);

	t=root;

	push(z, &sp, &stack);
	push(t, &sp, &stack);

	while(t!=z) {
		t=(NODE_XYCUT *)pop(&sp, &stack, &end_node);
		if(t==z) break;
  		free_zone[l++] = t;
  		if(SplitZone(buf, root, &t, document, x_hist_table, y_hist_table, para, 
               		&cut_direction, &updated_leftwidth, &updated_upperheight,
               		&free_hist, &h)==GO_ON) {

    			x=(NODE_XYCUT *) malloc (sizeof (NODE_XYCUT));
    			y=(NODE_XYCUT *) malloc (sizeof (NODE_XYCUT));

    			if(cut_direction == VERTICAL_CUT){
      				t->r=x; 
      				t->r->ul_x=t->ul_x+updated_leftwidth;
      				t->r->ul_y=t->ul_y;
      				t->r->width=t->width-updated_leftwidth;
      				t->r->height=t->height;
      				if(push(x, &sp, &stack)==-1)
					return (-1);
      				t->l=y; 
     				t->l->ul_x=t->ul_x;
      				t->l->ul_y=t->ul_y;
      				t->l->width=updated_leftwidth;
      				t->l->height=t->height;
      				if(push(y, &sp, &stack)==-1)
					return (-1);

    			}
    			else if(cut_direction == HORIZONTAL_CUT){
      				t->r=x; 
      				t->r->ul_x=t->ul_x;
      				t->r->ul_y=t->ul_y+updated_upperheight;
      				t->r->width=t->width;
      				t->r->height=t->height-updated_upperheight;
      				if(push(x, &sp, &stack)==-1)
					return (-1);
      				t->l=y; 
      				t->l->ul_x=t->ul_x;
      				t->l->ul_y=t->ul_y;
      				t->l->width=t->width;
      				t->l->height=updated_upperheight;
      				if(push(y, &sp, &stack)==-1) return (-1);
    			}

  		}

  		else {
  			t->r=z;
  			t->l=z;
  			t->info="This node is a leaf.";

  			/**--generate output in DAFS format----**/
  			if(t->width > 0 && t->height>0) {
    				dafs_zone=i_NewEntity(iParagraphT, dafs_doc);
    				dafs_box.x=t->ul_x;
			    	dafs_box.y=t->ul_y;
    				dafs_box.w=t->width;
    				dafs_box.h=t->height;
    				i_SetBox(dafs_zone, dafs_box);
 			}
  			/**----end of DAFS---------------------**/
  			PushLeaf(t, &leaf_sp, &leaf_stack);

  		}
	}

	/* Write dafs file */
	dafs_err = i_ReadImage(tif_filename, &imPtr);
	i_SetImage(dafs_doc, (iImagePtr) imPtr);
	dafs_err = i_WriteEntity(dafs_filename, dafs_doc, DAFSB);
	i_DisposeImage(imPtr);
	i_DisposeEntity(dafs_doc);
	i_ExitLib();

	free(stack);
	free(leaf_stack);
	for(i = 0; i < h; i++) {
 		free(free_hist[i]);
	}
	free(free_hist);
	for(i = 0; i < l; i++) {
 		free(free_zone[i]);
	}
	free(free_zone);
	free(z);

	return(0);
}

/* > */
 
/* < SPLIT IMAGE ZONE */
int SplitZone(char **buf, NODE_XYCUT *root, NODE_XYCUT **zone_p, DOCUMENT_IMAGE *document, 
              int **x_hist_table, int **y_hist_table, XYCUT_PARAM *para, 
              int *cut_direction_p, int *updated_leftwidth_p, int *updated_upperhight_p,
              int ***free_hist_p, int *h_p)
{
	int x_maxlength=0,y_maxlength=0;
	int x_midzeroes=0,y_midzeroes=0;
	int x_prev=0, y_prev=0;
	int x_cnt=0, y_cnt=0;
	int i;
	int split_flg;

	ZoneAdjust(buf, root, document, x_hist_table, y_hist_table, zone_p,
			free_hist_p, h_p);

	/*---------get horizontal widest zero vally length and its midpoint----*/
	for(i=0;i<(*zone_p)->width;i++){
		if((*zone_p)->x_histogram[i] > (*zone_p)->local_Tnx){
    			if(x_prev <= (*zone_p)->local_Tnx){
      				if(x_cnt > x_maxlength){
					x_maxlength = x_cnt;
					x_midzeroes = i-(x_maxlength>>1);
					x_cnt=0;
      				}
      				else
					x_cnt=0;
    			}
  		}
  		else if((*zone_p)->x_histogram[i] <= (*zone_p)->local_Tnx)
      			x_cnt++;
  		x_prev = (*zone_p)->x_histogram[i];   
	}

	/*---------get vertical widest zero vally length and its midpoint----*/
	for(i=0; i<(*zone_p)->height; i++){
  		if((*zone_p)->y_histogram[i] > (*zone_p)->local_Tny){
    			if(y_prev <= (*zone_p)->local_Tny){
      				if(y_cnt > y_maxlength){
					y_maxlength = y_cnt;
					y_midzeroes = i-(y_maxlength>>1);  
					y_cnt = 0;
      				}
      				else
					y_cnt = 0;
    			}
  		}
  		else if((*zone_p)->y_histogram[i] <= (*zone_p)->local_Tny)
      			y_cnt++;
  		y_prev=(*zone_p)->y_histogram[i];   
	}

	/*---------decide cutting position---------------------------------*/
	if(x_maxlength > para->Tcx || y_maxlength > para->Tcy){
  		split_flg=GO_ON;
  		if(x_maxlength > y_maxlength){
    			*updated_leftwidth_p=x_midzeroes;
    			*cut_direction_p = VERTICAL_CUT;
  		}
  		else {
    			*updated_upperhight_p=y_midzeroes;
    			*cut_direction_p = HORIZONTAL_CUT;
  		}
	}
	else
  		split_flg=STOP_HERE;

	return (split_flg);
}
 
/* > */

/* < GET HISTOGRAM OF CURRENT ZONE */
void GetHistogram(NODE_XYCUT **t, int **x_hist_table, int **y_hist_table,
		DOCUMENT_IMAGE *document)
{
  	int i,j;

  	(*t)->x_histogram = (int *) malloc ( (*t)->width * sizeof(int));
  	(*t)->y_histogram = (int *) malloc ( (*t)->height * sizeof(int));

  	for(i=0; i<(*t)->width; i++)
    		(*t)->x_histogram[i]=x_hist_table[(*t)->ul_y +
				(*t)->height - 1][(*t)->ul_x + i] - 
                         	x_hist_table[(*t)->ul_y][(*t)->ul_x + i];  

  		for(j=0; j<(*t)->height; j++)
    			(*t)->y_histogram[j]=
				y_hist_table[(*t)->ul_y + j][(*t)->ul_x + (*t)->width - 1] -
                         	y_hist_table[(*t)->ul_y + j][(*t)->ul_x];
}

/* > */

/* < PUSH subroutine */
int push(NODE_XYCUT *p, int *sp_p, NODE_XYCUT ***stack_p)
{
  	if((*sp_p)<MAX) {
    		(*stack_p)[(*sp_p)]=p;
    		(*sp_p)++;
    		return(0);}
  	else {
   		printf("error:stack full, can't push ");
   		return(-1);
	}
}

/* > */

/* < POP subroutine */
NODE_XYCUT * pop(int *sp_p, NODE_XYCUT ***stack_p, NODE_XYCUT **end_node_p)
{ 
  	if((*sp_p) > 0) { 
      		(*sp_p)--;
      		return ((*stack_p)[(*sp_p)]);
    	}
  	else {
    		printf("error: stack empty\n");
    		return((*end_node_p));
  	}
}
/* > */

/* < PUSH leaf subroutine */
int PushLeaf(NODE_XYCUT *p, int *leaf_sp_p, NODE_XYCUT ***leaf_stack_p)
{
  	if((*leaf_sp_p)<MAX) {
    		(*leaf_stack_p)[(*leaf_sp_p)]=p;
    		(*leaf_sp_p)++;
    		return(0);}
  	else {
    		printf("error:leaf stack full, can't push ");
   		return(-1);
	}
}

/* > */

/* < POP leaf subroutine */
NODE_XYCUT * PopLeaf(int *leaf_sp_p, NODE_XYCUT ***leaf_stack_p, NODE_XYCUT **end_node_p)
{ 
  	if((*leaf_sp_p) > 0) { 
      		(*leaf_sp_p)--;
      		return ((*leaf_stack_p)[(*leaf_sp_p)]);
    	}
  	else {
    		printf("error: leaf stack empty\n");
    		return((*end_node_p));
  	}
}

/* > */

/* < Free Job */
void FreeJobXycut(DOCUMENT_IMAGE *document, XYCUT_PARAM *para, char **buf,
			int **x_hist_table, int **y_hist_table)
{
  	int i;
  
  	for(i = 0; i < document->height; i++) {
    		free(buf[i]);
    		free(x_hist_table[i]);
    		free(y_hist_table[i]);
  	}

  	free(buf);
  	free(x_hist_table);
  	free(y_hist_table);

  	free(document->image);
  	free(document);
}
/* > */
